//var jojo    = require('jojo')

module.exports = require('joose').Class({
    
//    isa     : require('../Backend'),
    
    /*VERSION,*/
    
    use : [  ],
    
    
    has : {
        
        
    },
    
        
    methods : {
        
    }

})